"""The core of phial."""
import json
import logging
from typing import Callable, Dict, List, Optional

from slackclient import SlackClient  # type: ignore

from phial.commands import help_command
from phial.globals import _command_ctx_stack
from phial.scheduler import Schedule, ScheduledJob, Scheduler
from phial.types import PhialResponse
from phial.utils import parse_slack_output
from phial.wrappers import Attachment, Command, Message, Response


class Phial:
    """
    The Phial class acts as the main interface to Slack.

    It handles registraion and execution of user defined commands,
    as well as providing a wrapper around :obj:`slackclient.SlackClient`
    to make sending messages to Slack simpler.
    """

    #: Default configuration
    default_config = {
        'prefix': "!",
        'registerHelpCommand': True,
        'baseHelpText': "All available commands:",
        'autoReconnect': True,
    }

    def __init__(self,
                 token: str,
                 config: Dict = default_config,
                 logger: Optional[logging.Logger] = None) -> None:
        self.slack_client = SlackClient(token)
        self.commands: List[Command] = []
        self.config: Dict = dict(self.default_config)
        self.config.update(config)
        self.middleware_functions: List[Callable
                                        [[Message], Optional[Message]]] = []
        self.scheduler = Scheduler()
        self.fallback_func: Optional[Callable[[Message], PhialResponse]] = None
        if logger is None:
            logger = logging.getLogger(__name__)
            if not logger.hasHandlers():
                handler = logging.StreamHandler()
                formatter = logging.Formatter(fmt="%(asctime)s - %(message)s")
                handler.setFormatter(formatter)
                logger.addHandler(handler)
                logger.propagate = False
            logger.setLevel(logging.INFO)
        self.logger = logger
        self._register_standard_commands()

    def add_command(self,
                    pattern: str,
                    func: Callable[..., PhialResponse],
                    case_sensitive: bool = False,
                    help_text_override: Optional[str] = None) -> None:
        """
        Registers a command with the bot.

        This method can be used as a decorator via :meth:`command`

        :param pattern: The pattern that a :obj:`Message`'s text must
                        match for the command to be invoked.
        :param func: The fucntion to be executed when the command is
                     invoked
        :param case_sensitive: Whether the :code:`pattern` should
                               respect case sensitivity.

                               Defaults to False
        :param help_text_override: Text that should be used as a
                                   description of the command using
                                   the inbuilt help function.

                                   If not overriden the command's
                                   docstring will be used as the
                                   help text.

                                   Defaults to None

        :raises ValueError: If command with the same pattern is already
                            registered

        .. rubric:: Example

        ::

            def hello():
                return "world"
            bot.add_command('hello', world)


        Is the same as
        ::

            @bot.command('hello')
            def hello():
                return "world"

        """
        pattern = "{0}{1}".format(self.config["prefix"] if "prefix"
                                  in self.config else "", pattern)
        # Validate command does not already exist
        for existing_command in self.commands:
            if pattern == existing_command.pattern_string:
                raise ValueError('Command {0} already exists'
                                 .format(pattern.split("<")[0]))

        # Create and add command
        command = Command(pattern, func, case_sensitive, help_text_override)
        self.commands.append(command)
        self.logger.debug("Command {0} added"
                          .format(pattern))

    def command(self,
                pattern: str,
                case_sensitive: bool = False,
                help_text_override: Optional[str] = None) -> Callable:
        """
        Registers a command with the bot.

        This command is a decorator version of :meth:`add_command`

        :param pattern: The pattern that a :obj:`Message`'s text must
                        match for the command to be invoked.
        :param case_sensitive: Whether the :code:`pattern` should
                               respect case sensitivity.

                               Defaults to False
        :param help_text_override: Text that should be used as a
                                   description of the command using
                                   the inbuilt help function.

                                   If not overriden the command's
                                   docstring will be used as the
                                   help text.

                                   Defaults to None

        .. rubric:: Example

        ::

                @bot.command('hello')
                def hello():
                    return "world"

                @bot.command('caseSensitive', case_sensitive=True)
                def case_sensitive():
                    return "You typed caseSensitive"

        """
        def decorator(f: Callable) -> Callable:
            self.add_command(pattern, f, case_sensitive, help_text_override)
            return f
        return decorator

    def alias(self, pattern: str) -> Callable:
        """
        A decorator that is used to register an alias for a command.

        :param pattern: The pattern that a :obj:`Message`'s text must
                        match for the command to be invoked.

        .. rubric:: Example

        ::

            @bot.command('hello')
            @bot.alias('goodbye')
            def hello():
                return "world"
        """
        pattern = "{0}{1}".format(self.config["prefix"] if "prefix"
                                  in self.config else "", pattern)

        def decorator(f: Callable) -> Callable:
            if not hasattr(f, 'alias_patterns'):
                f.alias_patterns = []  # type: ignore
            f.alias_patterns.append(pattern)  # type: ignore
            return f
        return decorator

    def add_fallback_command(self,
                             func: Callable[[Message], PhialResponse]) -> None:
        """
        Add a fallback command.

        Registers a 'fallback' function to run when a user tries to execute a
        command that doesn't exist.

        This method can be used as a decorator via :meth:`fallback_command`

        :param func: The function to be executed when the user tries
                     to execute a command that doesn't exist

        .. rubric:: Example

        ::

            def error_handler(message: Message) -> str:
                return "Oops that command doesn't seem to exist"

            bot.add_fallback_command(error_handler)

        Is the same as
        ::

            @bot.fallback_command()
            def error_handler(message: Message) -> str:
                return "Oops that command doesn't seem to exist"

        """
        self.fallback_func = func

    def fallback_command(self) -> Callable:
        """
        A decorator to add a fallback command.

        See :meth:`add_fallback_command` for more information on
        fallback commands

        .. rubric:: Example

        ::

            @bot.fallback_command()
            def error_handler(message: Message) -> str:
                return "Oops that command doesn't seem to exist"
        """
        def decorator(f: Callable) -> Callable:
            self.add_fallback_command(f)
            return f
        return decorator

    def add_middleware(self,
                       func: Callable[[Message], Optional[Message]]) -> None:
        """
        Adds a middleware function to the bot.

        Middleware functions get passed every message the bot recieves from
        slack before the bot process the message itself. Returning :obj:`None`
        from a middleware function will prevent the bot from processing it.

        This method can be used as a decorator via :meth:`middleware`.

        :param middleware_func: The function to be added to the middleware
                                pipeline

        .. rubric :: Example

        ::

            def intercept(messaage):
                return message
            bot.add_middleware(intercept)

        Is the same as
        ::

            @bot.middleware()
            def intercept(messaage):
                return message

        """
        self.middleware_functions.append(func)
        self.logger.debug("Middleware {0} added"
                          .format(getattr(func, '__name__', repr(func))))

    def middleware(self) -> Callable:
        """
        A decorator to add a middleware function to the bot.

        See :meth:`add_middleware` for more information about middleware

        .. rubric:: Example

        ::

            @bot.middleware()
            def intercept(messaage):
                return message
        """
        def decorator(f: Callable) -> Callable:
            self.add_middleware(f)
            return f
        return decorator

    def add_scheduled(self,
                      schedule: Schedule,
                      func: Callable) -> None:
        """
        Adds a scheduled function to the bot.

        This method can be used as a decorator via :meth:`scheduled`.

        :param schedule: The schedule used to run the function
        :param scheduled_func: The function to be run in accordance to the
                               schedule

        .. rubric:: Example

        ::

            def scheduled_beep():
                bot.send_message(Response(text="Beep",
                                          channel="channel-id">))
            bot.add_scheduled(Schedule().every().day(), scheduled_beep)

        Is the same as

        ::

            @bot.scheduled(Schedule().every().day())
            def scheduled_beep():
                bot.send_message(Response(text="Beep",
                                          channel="channel-id">))
        """
        job = ScheduledJob(schedule, func)
        self.scheduler.add_job(job)
        self.logger.debug("Schedule {0} added"
                          .format(getattr(func, '__name__', repr(func))))

    def scheduled(self, schedule: Schedule) -> Callable:
        """
        A decorator that is used to register a scheduled function.

        See :meth:`add_scheduled` for more information on scheduled
        jobs.

        :param schedule: The schedule used to determine when the function
                         should be run

        .. rubric:: Example

        ::

            @bot.scheduled(Schedule().every().day())
            def scheduled_beep():
                bot.send_message(Response(text="Beep",
                                          channel="channel-id">))
        """
        def decorator(f: Callable) -> Callable:
            self.add_scheduled(schedule, f)
            return f
        return decorator

    def send_message(self, message: Response) -> None:
        """
        Sends a message to Slack.

        :param message: The message to be sent to Slack
        """
        api_method = ('chat.postEphemeral' if message.ephemeral
                      else 'chat.postMessage')

        if message.original_ts:
            self.slack_client.api_call(api_method,
                                       channel=message.channel,
                                       text=message.text,
                                       thread_ts=message.original_ts,
                                       attachments=json.dumps(
                                           message.attachments),
                                       user=message.user,
                                       as_user=True)
        else:
            self.slack_client.api_call(api_method,
                                       channel=message.channel,
                                       text=message.text,
                                       attachments=json.dumps(
                                           message.attachments),
                                       user=message.user,
                                       as_user=True)

    def send_reaction(self, response: Response) -> None:
        """
        Sends a reaction to a Slack Message.

        :param response: Response containing the reaction to be
                         sent to Slack
        """
        self.slack_client.api_call("reactions.add",
                                   channel=response.channel,
                                   timestamp=response.original_ts,
                                   name=response.reaction,
                                   as_user=True)

    def upload_attachment(self, attachment: Attachment) -> None:
        """
        Upload a file to Slack.

        :param attachment: The attachment to be uploaded to Slack
        """
        self.slack_client.api_call('files.upload',
                                   channels=attachment.channel,
                                   filename=attachment.filename,
                                   file=attachment.content)

    def _register_standard_commands(self) -> None:
        if ("registerHelpCommand" in self.config and self.
                config['registerHelpCommand']):
            # The command function has to be a lambda as we wish to delay
            # execution until all commands have been registered.
            self.add_command("help", lambda: help_command(self),
                             help_text_override="List all available commmands")

    def _send_response(self,
                       response: PhialResponse,
                       original_channel: str) -> None:
        if response is None:
            return  # Do nothing if command function returns nothing

        if isinstance(response, str):
            self.send_message(Response(text=response,
                                       channel=original_channel))

        elif not isinstance(response, Response) and not \
                isinstance(response, Attachment):
            raise ValueError('Only Response or Attachment objects can be '
                             'returned from command functions')

        if isinstance(response, Response):
            if response.original_ts and response.reaction and response.text:
                raise ValueError('Response objects with an original timestamp '
                                 'can only have one of the attributes: '
                                 'Reaction, Text')
            if response.original_ts and response.reaction:
                self.send_reaction(response)
            elif response.text or response.attachments:
                self.send_message(response)

        if isinstance(response, Attachment):
            self.upload_attachment(response)

    def _handle_message(self, message: Optional[Message]) -> None:
        if not message:
            return

        # Run middleware functions
        for func in self.middleware_functions:
            if message:
                message = func(message)

        # If message has been intercepted or is a bot message return early
        if not message or message.bot_id:
            return

        # If message should have a prefix but doesn't return early
        if "prefix" in self.config and self.config["prefix"] is not None \
                and not message.text.startswith(self.config["prefix"]):
            return

        # If message has not been intercepted continue with standard message
        # handling
        for command in self.commands:
            kwargs = command.pattern_matches(message)
            if kwargs is not None:
                try:
                    _command_ctx_stack.push(message)
                    response = command.func(**kwargs)
                    self._send_response(response, message.channel)
                    return
                finally:
                    _command_ctx_stack.pop()

        # If we are here then no commands have matched
        self.logger.warning("Command {0} not found".format(message.text))
        if self.fallback_func is not None:
            try:
                _command_ctx_stack.push(message)
                response = self.fallback_func(message)
                self._send_response(response, message.channel)
            finally:
                _command_ctx_stack.pop()

    def run(self) -> None:
        """
        Starts the bot.

        When called will start the bot listening to messages from Slack
        """
        auto_reconnect = self.config['autoReconnect']
        if not self.slack_client.rtm_connect(auto_reconnect=auto_reconnect,
                                             with_team_state=False):
            raise ValueError("Connection failed. Invalid Token or bot ID")

        self.logger.info("Phial connected and running!")
        while True:
            try:
                message = parse_slack_output(self.slack_client.rtm_read())
                self._handle_message(message)
                self.scheduler.run_pending()
            except Exception as e:
                self.logger.exception("Error {0}".format(e))



"""The classes related to scheduling of regular jobs in phial."""
from collections import namedtuple
from datetime import datetime, timedelta
from typing import Callable, List, Optional


class _Time(namedtuple("Time", ['hour', 'minute', 'second'])):
    """Represents a time of day.

    .. py:attribute:: hour

        The hour of day.

    .. py:attribute:: minute

        The minute of the hour.

    .. py:attribute:: second

        The second of the minute.

    """


class Schedule:
    """
    A schedule stores the relative time for something to happen.

    It can be used to compute when the next instance of an event
    should occur.
    """

    def __init__(self) -> None:
        self._days = 0
        self._at: Optional[_Time] = None
        self._hours = 0
        self._minutes = 0
        self._seconds = 0

    def every(self) -> 'Schedule':
        """
        Syntantic sugar to make schedule declaration more readable.

        Syntatic sugar to allow the declaration of schedules to be more
        like an English sentence.

        ::

            schedule = Schedule().every().day()
        """
        return self

    def day(self) -> 'Schedule':
        """
        Adds a day to the relative time till the next event.

        ::

            schedule = Schedule().every().day()
        """
        return self.days(1)

    def days(self, value: int) -> 'Schedule':
        """
        Set the days till the next instance of the event.

        Adds the specified number of days to the relative time till the next
        event.
        ::

            schedule = Schedule().every().days(2)

        :param value: The number of days to wait between events
        """
        self._days = value
        return self

    def at(self, hour: int, minute: int, second: int = 0) -> 'Schedule':
        """
        Specifies the time of day the next occurnce will happen.

        NOTE: 'at' can only be used with :meth:`day`.
        ::

            schedule = Schedule().every().day().at(12,00)

        :param hour: The hour of day the next event should happen,
                     when combined with the minute
        :param minute: The minute of day the next event should happen,
                       when combined with the hour
        :param second: The second of day the next event should
                       happen, when combined with the hour and
                       minute.
                       Defaults to 0
        """
        if self._hours or self._minutes:
            raise Exception("'at' can only be used on day(s)")
        if not self._days:
            raise Exception("'at' can only be used on day(s)")
        if self._at:
            raise Exception("'at' can only be set once")
        self._at = _Time(hour, minute, second)
        return self

    def hour(self) -> 'Schedule':
        """
        Adds an hour to the relative time till the next event.

        ::

            schedule = Schedule().every().hour()
        """
        return self.hours(1)

    def hours(self, value: int) -> 'Schedule':
        """
        Sets the hours till the next instance of the event.

        Adds the specified number of hours to the relative time till the next
        event.

        ::

            schedule = Schedule().every().hours(2)

        :param value: The number of hours to wait between events
        """
        self._hours = value
        return self

    def minute(self) -> 'Schedule':
        """
        Adds a minute to the relative time till the next event.

        ::

            schedule = Schedule().every().minute()
        """
        return self.minutes(1)

    def minutes(self, value: int) -> 'Schedule':
        """
        Sets the minutes till the next instance of the event.

        Adds the specified number of minutes to the relative time till the next
        event.

        ::

            schedule = Schedule().every().minutes(2)

        :param value: The number of minutes to wait between events
        """
        self._minutes = value
        return self

    def second(self) -> 'Schedule':
        """
        Adds a second to the relative time till the next event.

        ::

            schedule = Schedule().every().second()
        """
        return self.seconds(1)

    def seconds(self, value: int) -> 'Schedule':
        """
        Sets the seconds till the next instance of the event.

        Adds the specified number of seconds to the relative time till the next
        event.
        ::

            schedule = Schedule().every().seconds(2)

        :param value: The number of seconds to wait between events
        """
        self._seconds = value
        return self

    def get_next_run_time(self, last_run: datetime) -> datetime:
        """
        Get the next time the job should run.

        Calculates the next time to run, based on the last time the
        event was run.

        :param last_run: The last time the event happened

        :returns: A :obj:`datetime` of when the event should next happen
        """
        if self._at:
            next_run = last_run.replace(hour=self._at.hour,
                                        minute=self._at.minute,
                                        second=self._at.second,
                                        microsecond=0)
            if next_run <= datetime.now():
                next_run += timedelta(days=self._days)
            return next_run

        return last_run + timedelta(days=self._days,
                                    hours=self._hours,
                                    minutes=self._minutes,
                                    seconds=self._seconds)


class ScheduledJob:
    """A function with a schedule."""

    def __init__(self, schedule: Schedule, func: Callable) -> None:
        self.func = func
        self.schedule = schedule
        self.func = func
        self.next_run = self.schedule.get_next_run_time(datetime.now())

    def should_run(self) -> bool:
        """
        Checks whether the function needs to be run based on the schedule.

        :returns: A :obj:`bool` of whether or not to run
        """
        return self.next_run <= datetime.now()

    def run(self) -> None:
        """Runs the function and calculates + stores the next run time."""
        self.func()
        self.next_run = self.schedule.get_next_run_time(datetime.now())


class Scheduler:
    """A store for Scheduled Jobs."""

    def __init__(self) -> None:
        self.jobs: List[ScheduledJob] = []

    def add_job(self, job: ScheduledJob) -> None:
        """
        Adds a scheuled job to the scheduler.

        :param job: The job to be added to the scheduler
        """
        self.jobs.append(job)

    def run_pending(self) -> None:
        """
        Runs any pending scheduled jobs.

        Runs any ScheduledJobs in the store, where :code:`job.should_run()`
        returns true
        """
        jobs_to_run: List[ScheduledJob] = [job for job in self.jobs
                                           if job.should_run()]
        for job in jobs_to_run:
            job.run()



"""Contains models for phial to use."""

import re
from typing import IO, Callable, Dict, List, Optional, Pattern, Union

from phial.types import PhialResponse


class Response():
    r"""
    A response to be sent to Slack.

    When returned in a command function will send a message, or reaction to
    Slack depending on contents.

    :param channel: The Slack channel ID the response will be sent to
    :param text: The response contents
    :param original_ts: The timestamp of the original message. If populated
                        will put the text response in a thread
    :param reation: A valid slack emoji name. NOTE: will only work when
                    :code:`original_ts` is populated
    :param attachments: Any Slack
                        `Message Attachments <https://api.slack.com/docs/\
                        message-attachments#attachment_structure>`_
    :param ephemeral: Whether to send the message as an ephemeral message
    :param user: The user id to display the ephemeral message to

    .. rubric:: Examples

    The following would send a message to a slack channel when executed ::

        @slackbot.command('ping')
        def ping():
            return Response(text="Pong", channel='channel_id')

    The following would send a reply to a message in a thread ::

        @slackbot.command('hello')
        def hello():
            return Response(text="hi",
                            channel='channel_id',
                            original_ts='original_ts')

    The following would send a reaction to a message ::

        @slackbot.command('react')
        def react():
            return Response(reaction="x",
                            channel='channel_id',
                            original_ts='original_ts')
    """

    def __init__(self,
                 channel: str,
                 text: Optional[str] = None,
                 original_ts: Optional[str] = None,
                 reaction: Optional[str] = None,
                 ephemeral: bool = False,
                 user: Optional[str] = None,
                 attachments: Optional[Dict[str, Union[str,
                                                       int,
                                                       float,
                                                       bool]]] = None) -> None:
        self.channel = channel
        self.text = text
        self.original_ts = original_ts
        self.reaction = reaction
        self.ephemeral = ephemeral
        self.user = user
        self.attachments = attachments

    def __repr__(self) -> str:
        return "<Response: {0}>".format(self.text)

    def __eq__(self, other: object) -> bool:
        return self.__dict__ == other.__dict__


class Attachment():
    """
    A file to be uploaded to Slack.

    :param channel: The Slack channel ID the file will be sent to
    :param filename: The filename of the file
    :param content: The file to send to Slack. Open file
                    using open('<file>', 'rb')

    .. rubric:: Example

    ::

        Attachment('channel', 'file_name', open('file', 'rb'))
    """

    def __init__(self, channel: str, filename: str, content: IO) -> None:
        self.channel = channel
        self.filename = filename
        self.content = content

    def __repr__(self) -> str:
        return "<Attachment {0} in {1}>".format(self.filename, self.channel)


class Message():
    """
    A representation of a Slack message.

    :param text: The message contents
    :param channel: The Slack channel ID the message was sent from
    :param user: The user who sent the message
    :param timestamp: The message's timestamp
    :param team: The Team ID of the Slack workspace the message was
                 sent from
    :param bot_id: If the message was sent by a bot
                   the ID of that bot. Defaults to None.
    """

    def __init__(self,
                 text: str,
                 channel: str,
                 user: str,
                 timestamp: str,
                 team: Optional[str],
                 bot_id: Optional[str] = None) -> None:
        self.text = text
        self.channel = channel
        self.user = user
        self.timestamp = timestamp
        self.team = team
        self.bot_id = bot_id

    def __repr__(self) -> str:
        return "<Message: {0} in {1}:{2} at {3}>".format(self.text,
                                                         self.channel,
                                                         self.team,
                                                         self.timestamp)

    def __eq__(self, other: object) -> bool:
        return self.__dict__ == other.__dict__


class Command:
    """
    An action executable from Slack.

    :param pattern: A string that a Slack Message must match to trigger
                    execution
    :param func: The function that will be triggered when the command is
                  invoked
    :param case_sensitive: Whether the :code:`pattern` should enforce case
                           sensitivity
    :param help_text_override: Overrides the function's docstring in the
                               standard help command
    """

    def __init__(self,
                 pattern: str,
                 func: Callable[..., PhialResponse],
                 case_sensitive: bool = False,
                 help_text_override: Optional[str] = None):
        self.pattern_string = pattern
        self.pattern = self._build_pattern_regex(pattern, case_sensitive)
        self.alias_patterns = self._get_alias_patterns(func)
        self.func = func
        self.case_sensitive = case_sensitive
        self.help_text_override = help_text_override

    def __repr__(self) -> str:
        return "<Command: {0}>".format(self.pattern_string)

    def _get_alias_patterns(self, func: Callable) -> List[Pattern]:
        patterns: List[Pattern] = []
        if hasattr(func, 'alias_patterns'):
            for pattern in func.alias_patterns:  # type: ignore
                patterns.append(self._build_pattern_regex(pattern))
        return patterns

    @staticmethod
    def _build_pattern_regex(pattern: str,
                             case_sensitive: bool = False) -> Pattern[str]:
        command = re.sub(r'(<\w+>)', r'(\"?\1\"?)', pattern)
        command_regex = re.sub(r'(<\w+>)', r'(?P\1[^"]*)', command)
        return re.compile("^{}$".format(command_regex),
                          0 if case_sensitive else re.IGNORECASE)

    def pattern_matches(self, message: Message) -> Optional[Dict[str, str]]:
        """
        Check if message should invoke the command.

        Checks whether the text of a :obj:`Message` matches the command's
        pattern, or any of it's aliased patterns.
        """
        match = self.pattern.match(message.text)
        if match:
            return match.groupdict()

        # Only check aliases if main pattern does not match
        for alias in self.alias_patterns:
            match = alias.match(message.text)
            if match:
                return match.groupdict()

        return None

    @property
    def help_text(self) -> Optional[str]:
        """A description of the command's function."""
        if self.help_text_override is not None:
            return self.help_text_override
        return self.func.__doc__



"""Helper utilities for phial."""
import re
from typing import Dict, List, Optional

from phial.wrappers import Message


def parse_help_text(help_text: str) -> str:
    """Parse help text."""
    NEW_LINE_SEPERATOR = "<__NEW_LINE_SEPERATOR__>"

    # Strip excess whitespace
    help_text = help_text.strip()

    # Remove single new lines
    help_text = help_text.replace("\n", NEW_LINE_SEPERATOR)
    help_text = help_text.replace(NEW_LINE_SEPERATOR * 2, "\n")
    help_text = help_text.replace(NEW_LINE_SEPERATOR, "")

    # Remove extra spaces
    help_text = re.sub(r'(^[ \t]+|[ \t]+)', ' ', help_text, flags=re.M)

    return help_text


def parse_slack_output(slack_rtm_output: List[Dict]) -> Optional['Message']:
    """Parse Slack output."""
    output_list = slack_rtm_output
    if output_list and len(output_list) > 0:
        for output in output_list:
            if(output and 'text' in output):
                bot_id = None
                team = None
                if 'team' in output:
                    team = output['team']
                if 'bot_id' in output:
                    bot_id = output['bot_id']

                return Message(output['text'],
                               output['channel'],
                               output['user'],
                               output['ts'],
                               team,
                               bot_id)
    return None



"""phial Slack Bot."""
from phial.bot import Phial
from phial.globals import command
from phial.scheduler import Schedule
from phial.wrappers import Attachment, Message, Response

__version__ = "0.8.2"
__all__ = ['Phial', 'command', 'Response', 'Attachment', 'Schedule', 'Message']



"""Provides the globaly scoped parts of phial."""
from typing import cast

from werkzeug.local import LocalProxy, LocalStack

from phial.wrappers import Message


def _find_command() -> Message:
    """Gets the command from the context stack."""
    top = _command_ctx_stack.top
    if top is None:
        raise RuntimeError('Not in a context with a command')
    return cast(Message, top)


_command_ctx_stack = LocalStack()  # type: ignore
command: Message = cast(Message, LocalProxy(_find_command))



"""Shared types for phial."""
from typing import TYPE_CHECKING, Union

if TYPE_CHECKING:
    from phial.wrappers import Response, Attachment  # noqa

#: A union of all response types phial can use
PhialResponse = Union[None, str, 'Response', 'Attachment']



from setuptools import setup  # type: ignore
from setuptools.command.install import install  # type: ignore
import os
import codecs
import re
import sys

here = os.path.abspath(os.path.dirname(__file__))


def read(*parts: str) -> str:
    with codecs.open(os.path.join(here, *parts), 'r') as fp:
        return fp.read()


def find_version(*file_paths: str) -> str:
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


VERSION = find_version("phial", "__init__.py")


class VerifyVersionCommand(install):  # type: ignore
    """Custom command to verify that the git tag matches our version"""
    description = 'Verify that the git tag matches our version'

    def run(self) -> None:
        tag = os.getenv('CIRCLE_TAG')

        if tag != VERSION:
            info = "Git tag: {0} != phial version: {1}".format(tag,
                                                               VERSION)
            sys.exit(info)


setup(
    name='phial-slack',
    version=VERSION,
    url='https://github.com/sedders123/phial/',
    license='MIT',
    author='James Seden Smith',
    author_email='sedders123@gmail.com',
    description='A Slack bot framework',
    long_description=open('README.rst').read(),
    package_data={"phial": ["py.typed"]},
    packages=['phial'],
    include_package_data=True,
    zip_safe=False,
    platforms='any',
    python_requires='>=3.5',
    keywords=['Phial', 'Slack', 'bot', 'Slackbot'],
    install_requires=[
        'slackclient>=1.2.1',
        'Werkzeug>=0.14.1',
        'typing>=3.6.6'
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Typing :: Typed'
    ],
    cmdclass={
        'verify': VerifyVersionCommand,
    },
    project_urls={
        'CI: Circle': 'https://circleci.com/gh/sedders123/phial',
        'GitHub: Issues': 'https://github.com/sedders123/phial/issues',
        'Documentation': 'https://phial.readthedocs.io/en/develop/'
    },
)



"""Standard commands provided by phial."""
from typing import TYPE_CHECKING, cast

from phial.utils import parse_help_text

if TYPE_CHECKING:
    from phial import Phial  # noqa


def help_command(bot: 'Phial') -> str:
    """List all available commands."""
    help_text = cast(str, bot.config.get('baseHelpText', ""))
    if help_text:
        help_text += "\n"
    for command in bot.commands:
        command_doc = command.help_text
        if not command_doc:
            # If no help text default to blank string
            command_doc = ""
        command_help_text = parse_help_text(command_doc)
        help_text += "*{0}* - {1}\n".format(command.pattern_string,
                                            command_help_text)
    return help_text
